
static char help[] = "Solves a linear system in parallel with KSP.\n\n";

/*T
   Concepts: KSP^solving a Helmholtz equation
   Concepts: complex numbers;
   Concepts: Helmholtz equation
   Processors: n
T*/

/*
   Description: Solves a complex linear system in parallel with KSP.

   The model problem:
      Solve Helmholtz equation on the unit square: (0,1) x (0,1)
          -delta u - sigma1*u + i*sigma2*u = f, 
           where delta = Laplace operator
      Dirichlet b.c.'s on all sides
      Use the 2-D, five-point finite difference stencil.

   Compiling the code:
      This code uses the complex numbers version of PETSc, so configure
      must be run to enable this
*/

/* 
  Include "petscksp.h" so that we can use KSP solvers.  Note that this file
  automatically includes:
     petscsys.h       - base PETSc routines   petscvec.h - vectors
     petscmat.h - matrices
     petscis.h     - index sets            petscksp.h - Krylov subspace methods
     petscviewer.h - viewers               petscpc.h  - preconditioners
*/
#include <petscksp.h>
#include <petscao.h>
#include <petscis.h>

#undef __FUNCT__
#define __FUNCT__ "ModifySubMatrices"
PetscErrorCode ModifySubMatrices(PC pc,PetscInt nsub,const IS *row,const IS *col,Mat *submat,void *ctx)
{
  PetscMPIInt rank;
  
  MPI_Comm_rank(PETSC_COMM_WORLD,&rank);

  if (!rank) {
    printf("[0] submat[1] before SetValues: \n");
    MatView(submat[1],PETSC_VIEWER_STDOUT_SELF);
    
    PetscInt Ii, Ji;
    PetscScalar v;
    Ii= 1; Ji= 2; v= 0.0;
    MatSetValues(submat[1],1,&Ii,1,&Ji,&v,INSERT_VALUES);	/* this leads to error */
    MatAssemblyBegin(submat[1],MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(submat[1],MAT_FINAL_ASSEMBLY);
  }
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **args)
{
  Vec            u, x, xdd, b;      /* exact solution, approx solutions, RHS */
  Mat            A;            /* linear system matrix */
  KSP            ksp, kspdd;         /* linear solver context */
  PetscReal      norm;         /* norm of solution error */
  PetscInt       dim,i,j,Ii,Istart,Iend,n = 6,its, nsub, px, py, sx, sy, ipx, ipy,  ix, iy, *indices, numind;
  PetscErrorCode ierr;
  PetscScalar    v,none = -1.0,sigma2,pfive = 0.5,*xa;
  PetscReal      h2,sigma1 = 100.0;
  PetscBool      dd_gasm = PETSC_FALSE, dd_asm = PETSC_FALSE, flg = PETSC_FALSE, printx;
  char           dd_type[11];
  PetscScalar    a=1.0+PETSC_i;
  PC             pcdd;
  IS             *is_local, *is;
  PetscMPIInt    rank, size;

  PetscInitialize(&argc,&args,(char *)0,help);
#if !defined(PETSC_USE_COMPLEX)
  SETERRQ(PETSC_COMM_WORLD,PETSC_ERR_SUP,"This example requires complex scalars.");
#endif

  a=1.0+PETSC_i;
  printf("%G+%Gi\n",PetscRealPart(a),PetscImaginaryPart(a));

  ierr = PetscOptionsGetReal(PETSC_NULL,"-sigma1",&sigma1,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetInt(PETSC_NULL,"-n",&n,PETSC_NULL);CHKERRQ(ierr);
  dim = n*n;
  px= 1;  py= 1;
  ierr = PetscOptionsGetInt(PETSC_NULL,"-px",&px,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetInt(PETSC_NULL,"-py",&py,PETSC_NULL);CHKERRQ(ierr);
  sx= 1;  sy= 1;
  ierr = PetscOptionsGetInt(PETSC_NULL,"-sx",&sx,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetInt(PETSC_NULL,"-sy",&sy,PETSC_NULL);CHKERRQ(ierr);
  nsub= sx*sy;

  MPI_Comm_size(PETSC_COMM_WORLD,&size);
  if (size-px*py) {
    PetscPrintf(PETSC_COMM_WORLD,"Please use px*py processes! Quit\n");
    return 0;
  }

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
         Compute the matrix and right-hand-side vector that define
         the linear system, Ax = b.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Create parallel matrix, specifying both local and global dimensions.
     When using MatCreate(), the matrix format can be specified at
     runtime. Also, the parallel partitioning of the matrix is
     determined by PETSc at runtime.
  */

  MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
  ipx= rank/py; ipy= rank - ipx*py;
  PetscInt startpx, startpy, endpx, endpy;
  startpx= ipx*(n/px); startpy= ipy*(n/py);
  endpx= (ipx-px) ? (startpx + n/px) : n; 
  endpy= (ipy-py) ? (startpy + n/py) : n;
  PetscInt dimp;
  dimp= (endpx-startpx)*(endpy-startpy); 


  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);
  ierr = MatSetSizes(A,dimp,dimp,dim,dim);CHKERRQ(ierr);
  ierr = MatSetFromOptions(A);CHKERRQ(ierr);
  ierr = MatSetUp(A); CHKERRQ(ierr);

  /* 
     Currently, all PETSc parallel matrix formats are partitioned by
     contiguous chunks of rows across the processors.  Determine which
     rows of the matrix are locally owned. 
  */
  ierr = MatGetOwnershipRange(A,&Istart,&Iend);CHKERRQ(ierr);

  /*
     Create AO with natural global ordering mapped to PETSC's ordering: for Mat A
     We assume the domain is partitioned to px * py rectangles, and each 
     processor has one and the nodes on this processor are ordered as running
     y first
  */
  IS PetscIS; 
  AO ao;
  const PetscInt *PetscIndices;
  ierr= ISCreateStride(PETSC_COMM_SELF,dimp,Istart,1,&PetscIS);CHKERRQ(ierr);
  ierr= ISGetIndices(PetscIS,&PetscIndices);CHKERRQ(ierr);
  ierr= PetscMalloc(dimp*sizeof(PetscInt),&indices);CHKERRQ(ierr);
  numind= 0;
  for (ix = startpx; ix < endpx; ++ix)
  {
    for (iy = startpy; iy < endpy; ++iy)
    {
      indices[numind++]= ix*n + iy;
    }
  }
  ierr= AOCreateBasic(PETSC_COMM_WORLD,dimp,indices,PetscIndices,&ao);CHKERRQ(ierr);
  ierr= ISRestoreIndices(PetscIS,&PetscIndices);CHKERRQ(ierr);
  ISDestroy(&PetscIS);  PetscFree(indices);

  /* DEBUG: */
  /* ierr= AOView(ao,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr); */

  /*
     Set LocalToGlobal mapping.
  */
  numind= 0;  
  PetscInt xleft, xright, yleft, yright;
  ISLocalToGlobalMapping ltog;
  xleft=  (startpx > 0) ? startpx-1 : 0;   xright= (endpx < n) ? endpx+1 : n;
  yleft=  (startpy >0) ? startpy-1 : 0;  yright= (endpy < n) ? endpy+1: n;
  ierr= PetscMalloc((xright-xleft)*(yright-yleft)*sizeof(PetscInt),&indices);
  for (ix = xleft; ix < xright; ++ix)
  {
    for (iy = yleft; iy < yright; ++iy)
    {
      indices[numind++]= ix*n + iy;
    }
  }
  ierr= AOApplicationToPetsc(ao,(xright-xleft)*(yright-yleft),indices);
  ierr= ISLocalToGlobalMappingCreate(PETSC_COMM_WORLD,(xright-xleft)*(yright-yleft),indices,PETSC_COPY_VALUES,&ltog);CHKERRQ(ierr);
  ierr= PetscFree(indices);CHKERRQ(ierr);
  ierr= MatSetLocalToGlobalMapping(A,ltog,ltog);CHKERRQ(ierr);
 
  /* DEBUG: */
  /* ISLocalToGlobalMappingView(ltog,PETSC_VIEWER_STDOUT_SELF); */
  
  
  /* 
     Set matrix elements in parallel.
      - Each processor needs to insert only elements that it owns
        locally (but any non-local elements will be sent to the
        appropriate processor during matrix assembly). 
      - Always specify global rows and columns of matrix entries.
  */

  sigma2 = 10.0*PETSC_i;
  h2 = 1.0/((n+1)*(n+1));
  PetscInt dispx, dispy, row ,col;
  dispx= (xleft-startpx) ? 1 : 0;  dispy= (yleft-startpy) ? 1 : 0; 
  for (Ii=0; Ii<Iend-Istart; Ii++) { 
    v = -1.0; i = dispx + Ii/(endpy-startpy); j = dispy + (Ii % (endpy-startpy));  
    row =  i*(yright-yleft) + j;
    if (i>0) {
      col =  row - (yright-yleft); ierr = MatSetValuesLocal(A,1,&row,1,&col,&v,ADD_VALUES);CHKERRQ(ierr);}
    if (i<xright-xleft-1) {
      col = row + (yright-yleft); ierr = MatSetValuesLocal(A,1,&row,1,&col,&v,ADD_VALUES);CHKERRQ(ierr);}
    if (j>0) {
      col = row-1; ierr = MatSetValuesLocal(A,1,&row,1,&col,&v,ADD_VALUES);CHKERRQ(ierr);}
    if (j<yright-yleft-1) {
      col = row+1; ierr = MatSetValuesLocal(A,1,&row,1,&col,&v,ADD_VALUES);CHKERRQ(ierr);}
    v = 4.0 - sigma1*h2 + sigma2*h2;
    ierr = MatSetValuesLocal(A,1,&row,1,&row,&v,ADD_VALUES);CHKERRQ(ierr);
  }
  
  /* DEBUG: */
  PetscSynchronizedPrintf(PETSC_COMM_WORLD, "ipx= %d, ipy= %d, startpx= %d, endpx= %d, startpy= %d, endpy= %d\n", ipx,ipy,startpx,endpx,startpy,endpy);
  PetscSynchronizedPrintf(PETSC_COMM_WORLD, "xleft= %d, xright= %d, yleft= %d, yright= %d, dispx= %d, dispy= %d\n",xleft,xright,yleft,yright,dispx,dispy);
  PetscSynchronizedFlush(PETSC_COMM_WORLD);




  /* 
     Assemble matrix, using the 2-step process:
       MatAssemblyBegin(), MatAssemblyEnd()
     Computations can be done while messages are in transition
     by placing code between these two statements.
  */
  ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  MatView(A,PETSC_VIEWER_STDOUT_WORLD);

  /* 
     Create parallel vectors.
      - When using VecCreate(), VecSetSizes() and VecSetFromOptions(),
      we specify only the vector's global
        dimension; the parallel partitioning is determined at runtime. 
      - Note: We form 1 vector from scratch and then duplicate as needed.
  */
  ierr = VecCreate(PETSC_COMM_WORLD,&u);CHKERRQ(ierr);
  ierr = VecSetSizes(u,PETSC_DECIDE,dim);CHKERRQ(ierr);
  ierr = VecSetFromOptions(u);CHKERRQ(ierr);
  ierr = VecDuplicate(u,&b);CHKERRQ(ierr); 
  ierr = VecDuplicate(b,&x);CHKERRQ(ierr);
  ierr = VecDuplicate(b,&xdd);CHKERRQ(ierr);
 
  /* 
     Set exact solution; then compute right-hand-side vector.
  */
  
  ierr = VecSet(u,pfive);CHKERRQ(ierr);
  ierr = MatMult(A,u,b);CHKERRQ(ierr);

   /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                     Create the linear solver using PC(G)ASM.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  
  /* 
     Use the -dd_ prefix to set the options for the DD solver: -dd_ksp_type, etc. 
     Use prefix sub to set the subdomain solver options:
      -dd_sub_pc_type <pc_type> -dd_sub_ksp_type <ksp_type> -dd_sub_ksp_rtol 1.e-4 
  */
  ierr = KSPCreate(PETSC_COMM_WORLD,&kspdd);CHKERRQ(ierr);
  ierr = PetscObjectSetOptionsPrefix((PetscObject)kspdd, "dd_"); CHKERRQ(ierr);
  ierr = KSPSetOperators(kspdd,A,A,SAME_PRECONDITIONER);CHKERRQ(ierr);
  ierr = KSPGetPC(kspdd,&pcdd);CHKERRQ(ierr);
  ierr = PetscObjectSetOptionsPrefix((PetscObject)pcdd, "dd_"); CHKERRQ(ierr);
  dd_type[0] = 0;
  ierr = PetscOptionsGetString(PETSC_NULL, "-dd_type", dd_type, 10, &flg); CHKERRQ(ierr);
  ierr = PetscStrcmp(dd_type, "gasm", &dd_gasm); CHKERRQ(ierr);
  if(dd_gasm) {
    ierr = PCSetType(pcdd,PCGASM);CHKERRQ(ierr);
    PCGASMSetSortIndices(pcdd,PETSC_FALSE);
  }
  else {
    ierr = PetscStrcmp(dd_type, "asm", &dd_asm); CHKERRQ(ierr);
    if(dd_asm) {
      ierr = PCSetType(pcdd,PCASM);CHKERRQ(ierr);
      PCASMSetSortIndices(pcdd,PETSC_FALSE);
    }
    else {
      SETERRQ1(PETSC_COMM_WORLD, PETSC_ERR_SUP, "Unsupported -dd_type: %s", dd_type);
    }
  }

  /*
     Set the subdomains for PC(G)ASM
  */

  /* non-overlapping subdomains */
  PetscInt startsx, endsx, startsy, endsy;
  PetscMalloc(sx*sy*sizeof(IS),&is_local);
  for (i = 0; i < sx; ++i) {
    startsx= startpx + i*((endpx-startpx)/sx); 
    endsx= (i-sx+1) ? (startsx+(endpx-startpx)/sx) : endpx;
    for (j = 0; j < sy; ++j) {
      startsy= startpy + j*((endpy-startpy)/sy);
      endsy= (j-sy+1) ? (startsy+(endpy-startpy)/sy) : endpy;
      PetscMalloc((endsx-startsx)*(endsy-startsy)*sizeof(PetscInt),&indices);
      numind= 0;
      for (ix = startsx; ix < endsx; ++ix) {
        for (iy = startsy; iy < endsy; ++iy) {
	  indices[numind++]= ix*n + iy;
	}
      }
      AOApplicationToPetsc(ao,numind,indices);
      ISCreateGeneral(PETSC_COMM_SELF,numind,indices,PETSC_OWN_POINTER,is_local+i);
    } 
  } 
  

  /* overlapping subdomains */
  PetscInt expand= 1;
  PetscMalloc(sx*sy*sizeof(IS),&is);
  for (i = 0; i < sx; ++i) {
    startsx= startpx + i*((endpx-startpx)/sx) - expand;
    startsx= (startsx>0) ? startsx : 0;
    endsx= (i-sx+1) ? (startsx+(endpx-startpx)/sx) : endpx;
    endsx+= expand;  endsx= (endsx<n) ? endsx : n;
    for (j = 0; j < sy; ++j) {
      startsy= startpy + j*((endpy-startpy)/sy) - expand;
      startsy= (startsy>0) ? startsy : 0;
      endsy= (j-sy+1) ? (startsy+(endpy-startpy)/sy) : endpy;
      endsy+= expand;  endsy= (endsy<n)? endsy : n; 
      PetscMalloc((endsx-startsx)*(endsy-startsy)*sizeof(PetscInt),&indices);
      numind= 0;
      for (ix = startsx; ix < endsx; ++ix) {
        for (iy = startsy; iy < endsy; ++iy) {
	  indices[numind++]= ix*n + iy;
	}
      }
      AOApplicationToPetsc(ao,numind,indices);
      ISCreateGeneral(PETSC_COMM_SELF,numind,indices,PETSC_OWN_POINTER,is+i*sy+j);
      if (!rank && i*sy+j==1) {
	printf("[%d]: IS for overlapping subdomain 1 \n",rank);
	ISView(is[1],PETSC_VIEWER_STDOUT_SELF);
      } 
    } 
  }
  
  /* set PC(G)ASM with the subdomains */
  if(dd_gasm) {
    ierr= PCGASMSetSubdomains(pcdd,nsub,is_local, is);CHKERRQ(ierr);
  }
  else {
    ierr= PCASMSetLocalSubdomains(pcdd,nsub,is, is_local);CHKERRQ(ierr);
  }
  /* ierr= PCView(pcdd, PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr); */

  /*
     Modify submatrices
  */
  ierr= PCSetModifySubMatrices(pcdd,ModifySubMatrices,PETSC_NULL);CHKERRQ(ierr);

  /* 
     Set the DD solver options; use the -dd_ prefix: -dd_ksp_type, etc.  
   */
  ierr= KSPSetFromOptions(kspdd);CHKERRQ(ierr);
     


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                Create the benchmark linear solver.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /* 
     Create linear solver context
  */
  ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);


  /* 
     Set operators. Here the matrix that defines the linear system
     also serves as the preconditioning matrix.
  */
  ierr = KSPSetOperators(ksp,A,A,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr);


  /* 
     Use command-line options to set the benchmark KSP and PC options.
  */
#if 0
  /*
     Set type of KSP. 
  */  
  ierr= KSPSetType(ksp,KSPPREONLY);CHKERRQ(ierr);
  {
    PC pc;
    ierr= KSPGetPC(ksp,&pc);CHKERRQ(ierr);
    ierr= PCSetType(pc,PCLU);CHKERRQ(ierr);
    if (size==1) {
      ierr= PCFactorSetMatSolverPackage(pc,MATSOLVERSUPERLU);CHKERRQ(ierr);
    }
    else {
      ierr= PCFactorSetMatSolverPackage(pc,MATSOLVERSUPERLU_DIST);CHKERRQ(ierr);
    }
  }
#endif
  
  /* 
    Set runtime options, e.g.,
        -ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
  */
  ierr = KSPSetFromOptions(ksp);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
              Solve the linear system with two different solvers
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = KSPSolve(ksp,b,x);CHKERRQ(ierr);
  ierr = KSPSolve(kspdd,b,xdd);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                      Check solution x
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
      Print the first 3 entries of x; this demonstrates extraction of the
      real and imaginary components of the complex vector, x.
  */
  printx  = PETSC_FALSE;
  ierr = PetscOptionsGetBool(PETSC_NULL,"-print_x",&printx,PETSC_NULL);CHKERRQ(ierr);
  if (printx) {
    PetscInt nentries = 3, xlsize;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-print_x_entries", &nentries,PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"The first %D entries of x on each processor are:\n", nentries);CHKERRQ(ierr);
    ierr = VecGetLocalSize(x,&xlsize); CHKERRQ(ierr);
    nentries = PetscMin(xlsize, nentries);
    ierr = VecGetArray(x,&xa);CHKERRQ(ierr);
    for (i=0; i<nentries; i++){
      ierr = PetscSynchronizedPrintf(PETSC_COMM_WORLD,"[%D:%D]: x[%D] = %G + %G i\n",rank, size, i,PetscRealPart(xa[i]),PetscImaginaryPart(xa[i]));CHKERRQ(ierr);
  }
    ierr = PetscSynchronizedFlush(PETSC_COMM_WORLD); CHKERRQ(ierr);
    ierr = VecRestoreArray(x,&xa);CHKERRQ(ierr);
  }
  /* Check the error. */
  ierr = VecAXPY(x,none,u);CHKERRQ(ierr);
  ierr = VecNorm(x,NORM_2,&norm);CHKERRQ(ierr);
  ierr = KSPGetIterationNumber(ksp,&its);CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"Norm of error in x after %D iterations: %G\n",its, norm);CHKERRQ(ierr);


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                      Check solution xdd
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
      Print the first 3 entries of xdd; this demonstrates extraction of the
      real and imaginary components of the complex vector, x.
  */
  printx  = PETSC_FALSE;
  ierr = PetscOptionsGetBool(PETSC_NULL,"-dd_print_x",&printx,PETSC_NULL);CHKERRQ(ierr);
  if (printx) {
    PetscInt nentries = 3, xlsize;
    ierr = PetscOptionsGetInt(PETSC_NULL, "-dd_print_x_entries", &nentries,PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"The first %D entries of xdd on each processor are:\n", nentries);CHKERRQ(ierr);
    ierr = VecGetLocalSize(xdd,&xlsize); CHKERRQ(ierr);
    nentries = PetscMin(xlsize, nentries);
    ierr = VecGetArray(xdd,&xa);CHKERRQ(ierr);
    for (i=0; i<nentries; i++){
      ierr = PetscSynchronizedPrintf(PETSC_COMM_WORLD,"[%D:%D]: xdd[%D] = %G + %G i\n",rank, size, i,PetscRealPart(xa[i]),PetscImaginaryPart(xa[i]));CHKERRQ(ierr);
    }
    ierr = PetscSynchronizedFlush(PETSC_COMM_WORLD); CHKERRQ(ierr);
    ierr = VecRestoreArray(xdd,&xa);CHKERRQ(ierr);
  }
  /* Check the error. */
  ierr = VecAXPY(xdd,none,u);CHKERRQ(ierr);
  ierr = VecNorm(xdd,NORM_2,&norm);CHKERRQ(ierr);
  ierr = KSPGetIterationNumber(kspdd,&its);CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"Norm of error in xdd after %D iterations: %G\n",its, norm);CHKERRQ(ierr);


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                                 Clean up
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Free work space.  All PETSc objects should be destroyed when they
     are no longer needed.
  */
  ierr = KSPDestroy(&ksp);  CHKERRQ(ierr);
  ierr = KSPDestroy(&kspdd);CHKERRQ(ierr);
  ierr = VecDestroy(&u);CHKERRQ(ierr); 
  ierr = VecDestroy(&x);CHKERRQ(ierr);
  ierr = VecDestroy(&xdd);CHKERRQ(ierr);
  ierr = VecDestroy(&b);CHKERRQ(ierr); 
  ierr = MatDestroy(&A);CHKERRQ(ierr);
  ierr = PetscFinalize();
  return 0;
}
